#!/usr/bin/env python3
# Get current recentf (recently opened files) list from emacs.
# The list is sorted case insensitive.

# Note to developer: All error messages are to be presented to the user in one (well known) place.

import subprocess
import i3_sway_switch_window.display_error_message

def _emacs_recentf_list():
    """Return recentf (recently opened files) list from emacs.

    The list is in the same order as in emacs: from most recently accessed file, to least recently accessed file.
    """
    try:
        subprocess_result = subprocess.run(['emacs','--batch',
                                            '--eval',"(require 'server)",
                                            '--eval',"(mapc #'princ (read (server-eval-at \"server\" '(prin1-to-string (mapcar (lambda (buffer) (format \"%s\\n\" buffer)) recentf-list)))))"],
                                           capture_output=True, encoding="utf-8", check=True, timeout=5)
    except FileNotFoundError as exc:
        error_message = f"Process failed because the executable could not be found.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return []
    except subprocess.CalledProcessError as exc:
        error_message = f"Process failed because did not return a successful return code: {exc.returncode}\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return []
    except subprocess.TimeoutExpired as exc:
        error_message = f"Process timed out.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return []

    # print(subprocess_result.stdout)
    recentf_list = list(subprocess_result.stdout.split("\n"))
    recentf_list = [token for token in recentf_list if len(token) > 0]
    # recentf_list.sort(key=str.casefold)
    return recentf_list



__title__ = 'i3_sway_switch_window'
__description__ = 'In i3wm, or sway, replace current window with emacs buffer, browser tab or other window'
__url__ = 'https://github.com/johanwiden/i3-sway-switch-window'
__version__ = '0.2.0'
__author__ = 'Johan Widén'
__author_email__ = 'j.e.widen@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023 Johan Widén'



#!/usr/bin/env python3
# Get current list of buffers from emacs. Save buffer names into lists:
# - All interesting buffers (files and internal buffers)
# - All file buffers
# - All interesting internal buffers
# All lists are sorted case insensitive.

# Note to developer: All error messages are to be presented to the user in one (well known) place.

import subprocess
import i3_sway_switch_window.display_error_message

def _emacs_buffer_lists():
    """Return tree lists of the buffer names in emacs: All interesting buffers, file buffers, internal buffers.

    The lists are sorted case insensitive. Buffer names starting with a space character, are excluded.
    """
    try:
        # Command copied from https://blog.lambda.cx/posts/emacs-buffers-to-stdout/
        subprocess_result = subprocess.run(['emacs','--batch',
                                            '--eval',"(require 'server)",
                                            '--eval',"(mapc #'princ (read (server-eval-at \"server\" '(prin1-to-string (mapcar (lambda (buffer) (format \"%s\\n\" buffer)) (buffer-list))))))"],
                                           capture_output=True, encoding="utf-8", check=True, timeout=5)
    except FileNotFoundError as exc:
        error_message = f"Process failed because the executable could not be found.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return [[],[],[]]
    except subprocess.CalledProcessError as exc:
        error_message = f"Process failed because did not return a successful return code: {exc.returncode}\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return [[],[],[]]
    except subprocess.TimeoutExpired as exc:
        error_message = f"Process timed out.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
        return [[],[],[]]

    # print(subprocess_result.stdout)
    all_buffers = list(subprocess_result.stdout.split("\n"))
    interesting_buffers = [token for token in all_buffers if len(token) > 0 and token[0] != ' ']
    interesting_buffers.sort(key=str.casefold)
    file_buffers = [token for token in interesting_buffers if token[0] != '*']
    internal_buffers = [token for token in interesting_buffers if token[0] == '*']
    return [interesting_buffers, file_buffers, internal_buffers]



#!/usr/bin/env python3
# Get current list of web browser tab titles and URLs.
# The list is filtered for items containing string 'https://''
#   To support other tabs we have to ensure we use the same web browser the tab comes from.
# The list is sorted case insensitive.

# Note to developer: All error messages are to be presented to the user in one (well known) place.

import re
import subprocess
import i3_sway_switch_window.get_nyxt_title_url
import i3_sway_switch_window.display_error_message

def _browser_tab_list(get_urls_from_nyxt):
    """Return list of web browser tab titles and URLs.

    If get_urls_from_nyxt is True then also try to get titles and URLs from nyxt browser.
    The first column (tab ID), of each list item, is removed. The list is sorted case insensitive.
    """
    success = 1
    try:
        subprocess_result = subprocess.run(['bt','list'],
                                           capture_output=True, encoding="utf-8", check=True, timeout=5)
    except FileNotFoundError as exc:
        success = 0
        error_message = f"Process failed because the executable could not be found.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
    except subprocess.CalledProcessError as exc:
        success = 0
        error_message = f"Process failed because did not return a successful return code: {exc.returncode}\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
    except subprocess.TimeoutExpired as exc:
        success = 0
        error_message = f"Process timed out.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)

    if success:
        all_tabs = list(subprocess_result.stdout.split("\n"))
        all_tabs = [re.sub(r'^\S+\s', '', token) for token in all_tabs] # Remove tab ID column
        # Filter out internal URIs
        web_tabs = [token for token in all_tabs if 'https://' in token]
        if get_urls_from_nyxt:
            web_tabs = web_tabs + i3_sway_switch_window.get_nyxt_title_url._get_nyxt_title_url()
        # Make a sorted list, without duplicates
        set_res = set(web_tabs)
        web_tabs = sorted(list(set_res), key=str.casefold)
        return web_tabs
    else:
        return []



#!/usr/bin/env python3

import configparser
import os

import i3_sway_switch_window.__version__

parser = configparser.ConfigParser()

def _read_config_file(config_file_path):
    """Read and parse config file. config_file_path is uesed, if non empty.
    """
    file_path = ''
    if len(config_file_path) > 0:
        file_path = config_file_path
    else:
        config_path_suffix = os.path.join(i3_sway_switch_window.__version__.__title__, 'config.ini')
        possible_config_paths = []
        if 'APPDATA' in os.environ:
            path = os.path.join(os.environ['APPDATA'], config_path_suffix)
            possible_config_paths.append(path)

        if 'XDG_CONFIG_HOME' in os.environ:
            path = os.path.join(os.environ['XDG_CONFIG_HOME'], config_path_suffix)
            possible_config_paths.append(path)

        if 'HOME' in os.environ:
            path = os.path.join(os.environ['HOME'], ".config", config_path_suffix)
            possible_config_paths.append(path)

        if 'XDG_CONFIG_DIRS' in os.environ:
            path_list = os.environ['XDG_CONFIG_DIRS'].split(':')
            for path in path_list:
                possible_config_paths.append(os.path.join(path, config_path_suffix))

        for path in possible_config_paths:
            if os.path.exists(path):
                file_path = path
                break

    if len(file_path) == 0:
        i3_sway_switch_window.display_error_message._display_error_message('Config file ' + config_path_suffix + ' not found')
        return False
    elif not os.path.exists(file_path):
        i3_sway_switch_window.display_error_message._display_error_message('Config file ' + file_path + ' not found')
        return False

    return len(parser.read(file_path)) > 0

def _get_value(section, key):
    """Try to get a value from config file. If not found, the empty string is returned.
    """
    if parser.has_option(section, key):
        return parser[section][key]
    else:
        return ''



# Try to get a list of titles and URLs from the nyxt web browser.

import psutil
import re
import subprocess

import i3_sway_switch_window.display_error_message

timeout_seconds = 5

def _checkIfProcessRunning(processName):
    '''Check if there is any running process that contains the given name processName.
    '''
    # Iterate over the running process's
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if processName.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False;

def _get_nyxt_title_url():
    """Try to get a list of titles and URLs from nyxt.

    This function does nothing, if nyxt is not running.
    The list is returned formated similarly to brotab command "bt list", but without the first column.

    Requires tha emacs runs as a daemon, and that https://github.com/ag91/emacs-with-nyxt is loaded in
    emacs, or can be loaded with elisp function 'load-emacs-with-nyxt'.
    Requires that nyxt is configured to start a slynk server. In nyxt config.lisp add: (start-slynk)
    """
    if not _checkIfProcessRunning('nyxt'):
        return []

    success = 1
    try:
        # Try to load emacs-with-nyxt into emacs. Does nothing if function not defined, or if emacs-with-nyxt is
        # already loaded.
        subprocess_load_result = subprocess.run(['emacsclient', '-e',
                                                 "(if (fboundp 'load-emacs-with-nyxt) (load-emacs-with-nyxt))"],
                                                capture_output=True,
                                                encoding="utf-8", check=True, timeout=timeout_seconds)
        # Ensure that emacs has a slynk connection to nyxt.
        subprocess_connect_result = subprocess.run(['emacsclient', '-e',
                                                    '(unless (emacs-with-nyxt-connected-p) (emacs-with-nyxt-start-and-connect-to-nyxt))'],
                                                   capture_output=True,
                                                   encoding="utf-8", check=True, timeout=timeout_seconds)
        # Get list of buffer titles and URLs, from nyxt.
        subprocess_result = subprocess.run(['emacsclient', '-e',
                                            "(emacs-with-nyxt-send-sexps '(map 'list (lambda (el) (list (slot-value el 'title) (slot-value el 'url))) (buffer-list)))"],
                                           capture_output=True, encoding="utf-8", check=True, timeout=timeout_seconds)
    except FileNotFoundError as exc:
        success = 0
        error_message = f"Process failed because the executable could not be found.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
    except subprocess.CalledProcessError as exc:
        success = 0
        error_message = f"Process failed because did not return a successful return code: {exc.returncode}\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)
    except subprocess.TimeoutExpired as exc:
        success = 0
        error_message = f"Process timed out.\n{exc}"
        i3_sway_switch_window.display_error_message._display_error_message(error_message)

    if success:
        title_url_list = subprocess_result.stdout.split(")\\n")

        # Clean first and last item in list
        if len(title_url_list) > 0 and len(title_url_list[0]) > 2:
            item = title_url_list[0]
            title_url_list[0] = " " + item[2:]
            title_url_list[-1] = title_url_list[-1][:-5]
        # Filter out internal URIs
        title_url_list = [token for token in title_url_list if '#<QURI.URI:URI' not in token]
        # Create a list of 'tile  URL' formatted similarly to brotab command 'bt list'
        title_list = []
        for item in title_url_list:
            title_list.append(re.sub(r'^....([^\\]+).*$', r'\1', item) + '  '
                              + re.sub(r'^.*:URI-HTTPS (.*)>$', r'\1', item))

        # Make a sorted list, without duplicates
        set_res = set(title_list)
        title_list = sorted(list(set_res), key=str.casefold)
        return title_list
    else:
        return []



#!/usr/bin/env python3
# Display an error message in i3 or sway

import os
import subprocess

def _display_error_message(error_message):
    if os.environ.get('XDG_SESSION_TYPE') and os.environ['XDG_SESSION_TYPE'] == 'wayland':
        subprocess_result = subprocess.run(['swaynag','-t','error','-m',error_message], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
    else:
        subprocess_result = subprocess.run(['i3-nagbar','-t','error','-m',error_message], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)



#!/usr/bin/env python

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, 'i3_sway_switch_window', '__version__.py')) as f:
    exec(f.read(), about)

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/johanwiden/i3-sway-switch-window',
    classifiers=[
        # complete classifier list:
        # http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Environment :: X11 Applications',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Unix Shell',
        'Topic :: Desktop Environment :: Window Managers',
        ],
    project_urls = {
        "Bug Tracker": "https://github.com/johanwiden/i3-sway-switch-window/issues"
    },
    license='MIT License',
    packages=setuptools.find_packages(),
    # packages=['i3-switch-window'],
    install_requires=['i3ipc','psutil','stat','subprocess','tempfile'],
    entry_points = {
        'console_scripts': [
            'emacs_buffers = i3_sway_switch_window.main:cli_emacs_buffers',
            'emacs_recentf = i3_sway_switch_window.main:cli_emacs_recentf',
            'browser_tab = i3_sway_switch_window.main:cli_browser_tab',
            'wm_window_switch = i3_sway_switch_window.main:cli_wm_window_switch',
        ]
    },
)



#!/usr/bin/env python3
# Get current list of top level windows, from window manager.
# The list is sorted case insensitive.

# Note to developer: All error messages are to be presented to the user in one (well known) place.

import i3ipc

def _wm_window_list():
    """Return [focused_window.id, filtered_window_names].

    filtered_window_names is a list of top level window names, from window manager.
    The currently focused window is not included in the list.
    Each list element has the window id appended.
    The list is sorted case insensitive.
    """
    i3 = i3ipc.Connection()
    i3_tree = i3.get_tree()
    if not i3_tree:
        return []
    # Find currently focused window, in currently focused workspace
    focused_window = i3_tree.find_focused()
    if not focused_window:
        return []
    # print('Focused window %s %d is on workspace %s' %
    #       (focused_window.name, focused_window.id, focused_window.workspace().name))
    containers = i3_tree.leaves()
    # Do not include focused container
    filtered_containers = [c for c in containers if c.id != focused_window.id]
    filtered_names = [c.name + ' ' + str(c.id) for c in filtered_containers]
    if len(filtered_names) == 0:
        return []
    filtered_names.sort(key=str.casefold)
    return [focused_window.id, filtered_names]
